@@ -84,11 +84,7 @@ func TestNewMethodSet(t *testing.T) {
 	}
 
 	check := func(src string, methods []method, generic bool) {
-		pkg, err := pkgForMode("test", "package p;"+src, nil, 0)
-		if err != nil {
-			t.Errorf("%s: incorrect test case: %s", src, err)
-			return
-		}
+		pkg := mustTypecheck("test", "package p;"+src, nil)
 
 		scope := pkg.Scope()
 		if generic {
