@@ -301,6 +301,41 @@ func TestParseAddr(t *testing.T) {
 	}
 }
 
+func TestAddrFromSlice(t *testing.T) {
+	tests := []struct {
+		ip       []byte
+		wantAddr Addr
+		wantOK   bool
+	}{
+		{
+			ip:       []byte{10, 0, 0, 1},
+			wantAddr: mustIP("10.0.0.1"),
+			wantOK:   true,
+		},
+		{
+			ip:       []byte{0xfe, 0x80, 15: 0x01},
+			wantAddr: mustIP("fe80::01"),
+			wantOK:   true,
+		},
+		{
+			ip:       []byte{0, 1, 2},
+			wantAddr: Addr{},
+			wantOK:   false,
+		},
+		{
+			ip:       nil,
+			wantAddr: Addr{},
+			wantOK:   false,
+		},
+	}
+	for _, tt := range tests {
+		addr, ok := AddrFromSlice(tt.ip)
+		if ok != tt.wantOK || addr != tt.wantAddr {
+			t.Errorf("AddrFromSlice(%#v) = %#v, %v, want %#v, %v", tt.ip, addr, ok, tt.wantAddr, tt.wantOK)
+		}
+	}
+}
+
 func TestIPv4Constructors(t *testing.T) {
 	if AddrFrom4([4]byte{1, 2, 3, 4}) != MustParseAddr("1.2.3.4") {
 		t.Errorf("don't match")
@@ -559,9 +594,6 @@ func TestIPProperties(t *testing.T) {
 		private4b = mustIP("172.16.0.1")
 		private4c = mustIP("192.168.1.1")
 		private6  = mustIP("fd00::1")
-
-		unspecified4 = AddrFrom4([4]byte{})
-		unspecified6 = IPv6Unspecified()
 	)
 
 	tests := []struct {
@@ -686,12 +718,12 @@ func TestIPProperties(t *testing.T) {
 		},
 		{
 			name:        "unspecified v4Addr",
-			ip:          unspecified4,
+			ip:          IPv4Unspecified(),
 			unspecified: true,
 		},
 		{
 			name:        "unspecified v6Addr",
-			ip:          unspecified6,
+			ip:          IPv6Unspecified(),
 			unspecified: true,
 		},
 	}
@@ -752,6 +784,16 @@ func TestAddrWellKnown(t *testing.T) {
 			ip:   IPv6LinkLocalAllNodes(),
 			std:  net.IPv6linklocalallnodes,
 		},
+		{
+			name: "IPv6 link-local all routers",
+			ip:   IPv6LinkLocalAllRouters(),
+			std:  net.IPv6linklocalallrouters,
+		},
+		{
+			name: "IPv6 loopback",
+			ip:   IPv6Loopback(),
+			std:  net.IPv6loopback,
+		},
 		{
 			name: "IPv6 unspecified",
 			ip:   IPv6Unspecified(),
@@ -1835,6 +1877,8 @@ func TestNoAllocs(t *testing.T) {
 	test("ParseAddr/6", func() { sinkIP = panicIP(ParseAddr("::1")) })
 	test("MustParseAddr", func() { sinkIP = MustParseAddr("1.2.3.4") })
 	test("IPv6LinkLocalAllNodes", func() { sinkIP = IPv6LinkLocalAllNodes() })
+	test("IPv6LinkLocalAllRouters", func() { sinkIP = IPv6LinkLocalAllRouters() })
+	test("IPv6Loopback", func() { sinkIP = IPv6Loopback() })
 	test("IPv6Unspecified", func() { sinkIP = IPv6Unspecified() })
 
 	// IP methods
@@ -1908,10 +1952,10 @@ func TestAddrStringAllocs(t *testing.T) {
 		{"ipv4-in-ipv6", MustParseAddr("::ffff:192.168.1.1"), 1},
 		{"ipv4-in-ipv6+zone", MustParseAddr("::ffff:192.168.1.1%eth0"), 1},
 	}
-	isNooptBuilder := strings.HasSuffix(testenv.Builder(), "-noopt")
+	optimizationOff := testenv.OptimizationOff()
 	for _, tc := range tests {
 		t.Run(tc.name, func(t *testing.T) {
-			if isNooptBuilder && strings.HasPrefix(tc.name, "ipv4-in-ipv6") {
+			if optimizationOff && strings.HasPrefix(tc.name, "ipv4-in-ipv6") {
 				// Optimizations are required to remove some allocs.
 				t.Skipf("skipping on %v", testenv.Builder())
 			}
